/*-
 * Copyright (C)2000, 2001, 2002 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 * $Id: block.h,v 1.3 2007/03/08 23:30:10 babolo Exp $
 */

typedef float      float32_t;
typedef double     float64_t;
typedef char       bool;
typedef char      *Pointer;
typedef u_int16_t  LocationIndex;
typedef u_int32_t  Oid;
typedef u_int32_t  TransactionId;
typedef u_int32_t  CommandId;
typedef u_int32_t  StartUpID;
typedef u_int16_t  OffsetNumber;
#define MODULO(t,q,u) do { q = ((t) / (u)); if ((q) != 0) t -= ((q) * (u)); } while(0)

/************
 * page
 ************/

typedef struct {
    u_int32_t xlogid;   /* log file #, 0 based                 */
    u_int32_t xrecoff;  /* byte offset of location in log file */
} XLogRecPtr;

typedef struct {        /* line pointers            */
    unsigned lp_off:15  /* offset to start of tuple */
    ,        lp_flags:2 /* flags for tuple          */
    ,        lp_len:15  /* length of tuple          */
    ;
} ItemIdData;

#define LP_USED                 0x01    /* this line pointer is being used */
#define LP_DELETE               0x02    /* item is to be deleted           */
#define ItemIdGetLength(itemId) ((itemId)->lp_len)
#define ItemIdGetOffset(itemId) ((itemId)->lp_off)
#define ItemIdGetFlags(itemId)  ((itemId)->lp_flags)
#define ItemIdIsUsed(itemId)    (((itemId)->lp_flags & LP_USED) != 0)
#define ItemIdDeleted(itemId)   (((itemId)->lp_flags & LP_DELETE) != 0)

typedef struct {
    XLogRecPtr    pd_lsn;     /* LSN: next byte after last byte of xlog    */
    StartUpID     pd_sui;     /* SUI of last changes (currently it's       */
    LocationIndex pd_lower;   /* offset to start of free space             */
    LocationIndex pd_upper;   /* offset to end of free space               */
    LocationIndex pd_special; /* offset to start of special space          */
    u_int16_t     pd_pagesize_version;
    ItemIdData    pd_linp[1]; /* beginning of line pointer array           */
} PageHeaderData;

#define PG_PAGE_LAYOUT_VERSION         1
#define SizeOfPageHeaderData           ((long)&((PageHeaderData*)0)->pd_linp[0])
#define PageIsEmpty(page)              (((PageHeaderData*)(page))->pd_lower <= SizeOfPageHeaderData)
#define PageIsNew(page)                (((PageHeaderData*)(page))->pd_upper == 0)
#define PageGetContents(page)          ((char*)(&((PageHeaderData*)(page))->pd_linp[0]))
#define PageGetItem(page, itemId)      (Pointer)(((char*)(page)) + (itemId)->lp_off)
#define PageGetPageLayoutVersion(page) (((PageHeaderData*)(page))->pd_pagesize_version & 0x00FF)
#define PageGetPageSize(page) \
    ((size_t)(((PageHeaderData*)(page))->pd_pagesize_version & (u_int16_t)0xFF00))
#define PageGetItemId(page, offsetNumber) \
    ((ItemIdData*)(&((PageHeaderData*)(page))->pd_linp[(offsetNumber) - 1]))
#define PageGetMaxOffsetNumber(page)                                                      \
    ( ((PageHeaderData*)(page))->pd_lower <= SizeOfPageHeaderData                         \
    ? 0                                                                                   \
    : ((((PageHeaderData*)(page))->pd_lower - SizeOfPageHeaderData) / sizeof(ItemIdData)) \
    )
#define InvalidOffsetNumber             ((OffsetNumber)0)
#define FirstOffsetNumber               ((OffsetNumber)1)

/************
 * tuple
 ************/

typedef struct {
    u_int16_t bi_hi;
    u_int16_t bi_lo;
} BlockIdData;

typedef struct {
    BlockIdData  ip_blkid;
    OffsetNumber ip_posid;
} ItemPointerData;

typedef struct {
    TransactionId   t_xmin;     /* inserting xact ID                     */
    union {
        CommandId     t_cmin;   /* inserting command ID                  */
        TransactionId t_xmax;   /* deleting xact ID                      */
    }               t_field2;
    union {
        CommandId     t_cmax;   /* deleting command ID                   */
        TransactionId t_xvac;   /* VACUUM FULL xact ID                   */
    }               t_field3;
    ItemPointerData t_ctid;     /* current TID of this or newer tuple    */
    int16_t         t_natts;    /* number of attributes */
    u_int16_t       t_infomask; /* various flag bits, see below          */
    u_int8_t        t_hoff;     /* sizeof header incl. bitmap, padding   */
    /* ^ - 23 bytes - ^ */
    u_int8_t           t_bits[1];  /* bitmap of NULLs -- VARIABLE LENGTH */
    /* MORE DATA FOLLOWS AT END OF STRUCT */
} HeapTupleHeaderData;

#define HEAP_HASNULL           0x0001  /* has null attribute(s)                       */
#define HEAP_HASVARWIDTH       0x0002  /* has variable-width attribute(s)             */
#define HEAP_HASEXTERNAL       0x0004  /* has external stored attribute(s)            */
#define HEAP_HASCOMPRESSED     0x0008  /* has compressed stored attribute(s)          */
#define HEAP_HASEXTENDED       0x000C  /* the two above combined                      */
#define HEAP_HASOID            0x0010  /* has an object-id field                      */
                        /* bit 0x0020 is presently unused */
#define HEAP_XMAX_IS_XMIN      0x0040  /* created and deleted in the same transaction */
#define HEAP_XMAX_UNLOGGED     0x0080  /* to lock tuple for update without logging    */
#define HEAP_XMIN_COMMITTED    0x0100  /* t_xmin committed                            */
#define HEAP_XMIN_INVALID      0x0200  /* t_xmin invalid/aborted                      */
#define HEAP_XMAX_COMMITTED    0x0400  /* t_xmax committed                            */
#define HEAP_XMAX_INVALID      0x0800  /* t_xmax invalid/aborted                      */
#define HEAP_MARKED_FOR_UPDATE 0x1000  /* marked for UPDATE                           */
#define HEAP_UPDATED           0x2000  /* this is UPDATEd version of row              */
#define HEAP_MOVED_OFF         0x4000  /* moved to another place by VACUUM FULL       */
#define HEAP_MOVED_IN          0x8000  /* moved from another place by VACUUM FULL     */
#define HEAP_XACT_MASK         0xFFC0  /* visibility-related bits                     */
#define HEAP_MOVED (HEAP_MOVED_OFF | HEAP_MOVED_IN)

#define BITMAPLEN(NATTS)     (((int)(NATTS) + 7) / 8)
#define InvalidOid           ((Oid)0)
#define FirstCommandId       ((CommandId)0)

/************
 * varlena
 ************/

typedef struct {
    int32_t va_header;             /* External/compressed storage */
    /* flags and item size */
    union {
        struct {
            int32_t va_rawsize;    /* Plain data size             */
            char    va_data[1];    /* Compressed data             */
        }     va_compressed;       /* Compressed stored attribute */
        struct {
            int32_t va_rawsize;    /* Plain data size             */
            int32_t va_extsize;    /* External saved size         */
            Oid     va_valueid;    /* Unique identifier of value  */
            Oid     va_toastrelid; /* RelID where to find chunks  */
        }     va_external;         /* External stored attribute   */
        char  va_data[1];          /* Plain stored attribute      */
    }       va_content;
} varattrib;

struct varlena {
    int32_t vl_len;
    char    vl_dat[1];
};

#define VARHDRSZ       ((int32_t)sizeof(int32_t))
#define VARDATA(PTR)   (((varattrib*)(PTR))->va_content.va_data)

/************
 * float
 ************/
#define DBL_DIG 15
#define FLT_DIG 6

/************
 * numeric
 ************/
/*
 * The Numeric data type stored in the database
 *
 * NOTE: by convention, values in the packed form have been stripped of
 * all leading and trailing zero digits (where a "digit" is of base NBASE).
 * In particular, if the value is zero, there will be no digits at all!
 * The weight is arbitrary in that case, but we normally set it to zero.
 */
typedef struct {
    int32_t   varlen;         /* Variable size (std varlena header)    */
    int16_t   n_weight;       /* Weight of 1st digit                   */
    u_int16_t n_sign_dscale;  /* Sign + display scale                  */
    char      n_data[1];      /* Digits (really array of NumericDigit) */
} NumericData;

#define NUMERIC_HDRSZ       (sizeof(int32_t) + sizeof(int16_t) + sizeof(u_int16_t))
#define NUMERIC_SIGN_MASK   0xC000
#define NUMERIC_POS         0x0000
#define NUMERIC_NEG         0x4000
#define NUMERIC_NAN         0xC000
#define NUMERIC_DSCALE_MASK 0x3FFF
#define NUMERIC_SIGN(n)     ((n)->n_sign_dscale & NUMERIC_SIGN_MASK)
#define NUMERIC_DSCALE(n)   ((n)->n_sign_dscale & NUMERIC_DSCALE_MASK)
#define NUMERIC_IS_NAN(n)   (NUMERIC_SIGN(n) != NUMERIC_POS && NUMERIC_SIGN(n) != NUMERIC_NEG)
#define NBASE(b,x)          ((b) == (x))
#define PG_DIGITS(b)        ((NBASE((b), 10000)) ? 2 : 1)
#define DEC_DIGITS(b)       ((NBASE((b), 10000)) ? 4 : (NBASE((b), 100)) ? 2 : 1)
#define DIGIT(b,d,n)        (((b) == 10000) \
                          ? (           (((u_int16_t*)&(d))[(n)])) \
                          : ((u_int16_t)(((u_int8_t*)&(d))[(n)])))

/************
 * name
 ************/
#define NAMEDATALEN 64

typedef union {
    char data[NAMEDATALEN];
    int  alignmentDummy;
} NameData;

typedef NameData *Name;
#define NameStr(name)   ((name).data)

/************
 * inet
 ************/
typedef struct varlena inet;

typedef struct {
    unsigned char a;
    unsigned char b;
    unsigned char c;
    unsigned char d;
    unsigned char e;
    unsigned char f;
} macaddr;

typedef struct {
    unsigned char family;
    unsigned char bits;
    unsigned char type;
    unsigned char ip_addr[16];      /* 128 bits of address */
} inet_struct;

#define PGSQL_AF_INET  (AF_INET + 0)
#define PGSQL_AF_INET6 (AF_INET + 1)
#define ip_family(inetptr) (((inet_struct*)VARDATA(inetptr))->family)
#define ip_bits(inetptr)   (((inet_struct*)VARDATA(inetptr))->bits)
#define ip_type(inetptr)   (((inet_struct*)VARDATA(inetptr))->type)
#define ip_addr(inetptr)   (((inet_struct*)VARDATA(inetptr))->ip_addr)

/************
 * array
 ************/
#define MAXDIM 6

typedef struct {
    int indx[MAXDIM];
} IntArray;

typedef struct {
    int32_t size;        /* total array size (varlena requirement) */
    int     ndim;        /* # of dimensions                        */
    int     flags;       /* implementation flags                   */
    Oid     elemtype;    /* element type OID                       */
} ArrayType;
